/***************************************************************
 * Name:      engine.cpp
 * Author:    David Vachulka (arch_dvx@users.sourceforge.net)
 * Copyright: 2013
 * License:   GPL3
 **************************************************************/

#include <wx/datetime.h>

#include "engine.h"
#include "sqlite3.h"

Engine::~Engine()
{
    try
    {
        m_db.Close();
    }
    catch(const wxSQLite3Exception& e)
    {
        wxLogError(wxString::Format(_T("Close error: %s"), e.GetMessage().c_str()));
    }
}

bool Engine::Open(const wxString &filename)
{
    int err = SQLITE_OK;
    wxString errStr=wxEmptyString;
    try
    {
        m_db.Open(filename);
    }
    catch(const wxSQLite3Exception &e)
    {
        err = e.GetErrorCode();
        errStr << e.GetMessage();
    }
    if(err == SQLITE_OK)
    {
        //Try if db selectable
        try
        {
            m_db.ExecuteScalar("select count(*) from  sqlite_master where type='table'");
        }
        catch(const wxSQLite3Exception &e)
        {
            err = e.GetExtendedErrorCode();
            errStr = e.GetMessage();
        }
        CreateEventsTable();
        CreateRecurrenceTable();
        return true;
    }
    return false;
}

void Engine::LoadData()
{
    try
    {
        static const char sql[] = "select * from EVENTS";
        wxSQLite3ResultSet q = m_db.ExecuteQuery(sql);
        for( ; q.NextRow(); )
        {
            m_events.push_back(Event(q.GetInt(_T("EVENTID")), q.GetString(_T("EVENT")), q.GetString(_T("DESCRIPTION")),
                     q.GetDateTime(_T("EVENTDATE")), q.GetInt64(_T("ACTUALREMINDER")),
                     q.GetBool(_T("REMINDED"))));
        }
        q.Finalize();
        static const char sql1[] = "select * from RECURRENCE";
        q = m_db.ExecuteQuery(sql1);
        for( ; q.NextRow(); )
        {
            m_recurrences.push_back(Recurrence(q.GetInt(_T("RECURRENCEID")), q.GetInt(_T("TYPE")), q.GetInt(_T("DAYS")), q.GetInt(_T("MONTHS")), q.GetInt(_T("EVENTID"))));
        }
        q.Finalize();
    }
    catch(const wxSQLite3Exception& e)
    {
        wxLogError(wxString::Format(_("Error in LoadData: %s"), e.GetMessage().c_str()));
    }
}

void Engine::CreateEventsTable()
{
    try
    {
        bool exists = m_db.TableExists(_T("EVENTS"));
        if(!exists)
        {
            m_db.ExecuteUpdate(_T("create table EVENTS(EVENTID integer primary key, \
                                  EVENT text not null, DESCRIPTION text, EVENTDATE text not null, \
                                  REMINDER integer not null, ACTUALREMINDER integer not null, \
                                  REMINDED integer not null);"));
            exists = m_db.TableExists(_T("EVENTS"));
            wxASSERT(exists);
        }
    }
    catch(const wxSQLite3Exception& e)
    {
        wxLogError(wxString::Format(_("Error in CreateEventsTable: %s"), e.GetMessage().c_str()));
    }
}

void Engine::CreateRecurrenceTable()
{
    try
    {
        bool exists = m_db.TableExists(_T("RECURRENCE"));
        if(!exists)
        {
            m_db.ExecuteUpdate(_T("create table RECURRENCE(RECURRENCEID integer primary key, \
                                  TYPE integer not null, DAYS integer not null, MONTHS integer not null, \
                                  EVENTID integer not null);"));
            exists = m_db.TableExists(_T("RECURRENCE"));
            wxASSERT(exists);
        }
    }
    catch(const wxSQLite3Exception& e)
    {
        wxLogError(wxString::Format(_("Error in CreateRecurrenceTable: %s"), e.GetMessage().c_str()));
    }
}

Event Engine::AddEvent(const wxString &event, const wxString &descr, const wxDateTime &date, const wxLongLong &reminder)
{
    try
    {
        static const char sql[] =
            "insert into EVENTS ("
            "EVENT, DESCRIPTION, EVENTDATE, REMINDER, ACTUALREMINDER, REMINDED)"
            "values (?, ?, ?, ?, ?, ?)";
        wxSQLite3Statement st = m_db.PrepareStatement(sql);
        st.Bind(1, event);
        st.Bind(2, descr);
        st.BindDateTime(3, date);
        st.Bind(4, reminder);
        st.Bind(5, reminder);
        st.BindBool(6, false);
        st.ExecuteUpdate();
        long id = m_db.GetLastRowId().ToLong();
        st.Finalize();
        Event ev(id, event, descr, date, reminder, false);
        m_events.push_back(ev);
        return ev;
    }
    catch(const wxSQLite3Exception& e)
    {
        wxLogError(wxString::Format(_("Error in AddEvent: %s"), e.GetMessage().c_str()));
    }
    return Event();
}

void Engine::EditEvent(long id, const wxString &event, const wxString &descr, const wxDateTime &date, const wxLongLong &reminder)
{
    try
    {
        static const char sql[] =
            "update EVENTS "
            "set EVENT = ?, DESCRIPTION = ?, EVENTDATE = ?, REMINDER = ?, ACTUALREMINDER = ?"
            "where EVENTID = ?";
        wxSQLite3Statement st = m_db.PrepareStatement(sql);
        st.Bind(1, event);
        st.Bind(2, descr);
        st.BindDateTime(3, date);
        st.Bind(4, reminder);
        st.Bind(5, reminder);
        st.Bind(6, wxLongLong(id));
        st.ExecuteUpdate();
        st.Finalize();
        size_t index = GetEventIndexByID(id);
        if(index != -1)
        {
            m_events[index].SetEvent(event);
            m_events[index].SetDescr(descr);
            m_events[index].SetDate(date);
            m_events[index].SetReminder(reminder);
        }
    }
    catch(const wxSQLite3Exception& e)
    {
        wxLogError(wxString::Format(_("Error in EditEvent: %s"), e.GetMessage().c_str()));
    }
}

void Engine::EditEvent(long id, const wxLongLong &reminder)
{
    try
    {
        static const char sql[] =
            "update EVENTS "
            "set ACTUALREMINDER = ?, REMINDED = ?"
            "where EVENTID = ?";
        wxSQLite3Statement st = m_db.PrepareStatement(sql);
        st.Bind(1, reminder);
        st.Bind(2, false);
        st.Bind(3, wxLongLong(id));
        st.ExecuteUpdate();
        st.Finalize();
        size_t index = GetEventIndexByID(id);
        if(index != -1)
        {
            m_events[index].SetReminder(reminder);
            m_events[index].SetReminded(false);
        }
    }
    catch(const wxSQLite3Exception& e)
    {
        wxLogError(wxString::Format(_("Error in EditEvent: %s"), e.GetMessage().c_str()));
    }
}

void Engine::RemoveEvent(long id)
{
    try
    {
        static const char* sql[] =
        {
            "delete from EVENTS where EVENTID = ?",
            "delete from RECURRENCE where EVENTID = ?",
            0
        };
        for(int i = 0; sql[i]; ++i)
        {
            wxSQLite3Statement st = m_db.PrepareStatement(sql[i]);
            st.Bind(1, wxLongLong(id));
            st.ExecuteUpdate();
            st.Finalize();
        }
        std::vector<Event>::iterator i;
        for(i = m_events.begin(); i != m_events.end(); ++i)
        {
            if((*i).GetID() == id)
            {
                m_events.erase(i);
                break;
            }
        }
    }
    catch(const wxSQLite3Exception& e)
    {
        wxLogError(wxString::Format(_("Error in RemoveEvent: %s"), e.GetMessage().c_str()));
    }
}

void Engine::RecurrentEvent(long id)
{
    size_t index = GetEventIndexByID(id);
    if(index != -1 && HasRecurrence(id))
    {
        Recurrence rec = GetRecurrenceByEventID(id);
        wxDateTime date = m_events[index].GetDate()+wxDateSpan::Months(rec.GetMonths())+wxDateSpan::Days(rec.GetDays());
        wxDateTime now = wxDateTime::Now();
        while(date < now)
        {
            date = date+wxDateSpan::Months(rec.GetMonths())+wxDateSpan::Days(rec.GetDays());
        }
        try
        {
            wxLongLong reminder = 0;
            wxSQLite3ResultSet q = m_db.ExecuteQuery(wxString::Format(_T("select REMINDER from EVENTS where EVENTID = '%ld';"), id));
            if(q.NextRow())
            {
                reminder = q.GetInt64(0);
            }
            q.Finalize();
            static const char sql[] =
                "update EVENTS "
                "set EVENTDATE = ?, REMINDER = ?, ACTUALREMINDER = ?, REMINDED = ?"
                "where EVENTID = ?";
            wxSQLite3Statement st = m_db.PrepareStatement(sql);
            st.BindDateTime(1, date);
            st.Bind(2, reminder);
            st.Bind(3, reminder);
            st.BindBool(4, false);
            st.Bind(5, wxLongLong(id));
            st.ExecuteUpdate();
            st.Finalize();
            size_t index = GetEventIndexByID(id);
            if(index != -1)
            {
                m_events[index].SetDate(date);
                m_events[index].SetReminder(reminder);
                m_events[index].SetReminded(false);
            }
        }
        catch(const wxSQLite3Exception& e)
        {
            wxLogError(wxString::Format(_("Error in RecurrentEvent: %s"), e.GetMessage().c_str()));
        }
    }
}

wxString Engine::GetItemText(long item, long column) const
{
    if(column == 0)
        return m_events[item].GetEvent();
    else if(column == 1)
        return m_events[item].GetDate().Format(_T("%x %X"));
    else if(column == 2)
        return m_events[item].GetReminderDate().Format(_T("%x %X"));
    return wxGetEmptyString();
}

Event Engine::GetEvent(long index)
{
    if(index <0 || index >= m_events.size())
        return Event();
    return m_events[index];
}

Event Engine::GetEventByID(long id)
{
    for(size_t i=0; i<m_events.size(); i++)
    {
        if(m_events[i].GetID() == id)
            return m_events[i];
    }
    return Event();
}

size_t Engine::GetEventIndexByID(long id)
{
    for(size_t i=0; i<m_events.size(); i++)
    {
        if(m_events[i].GetID() == id)
            return i;
    }
    return -1;
}

std::vector<Event> Engine::GetEventsByReminder(wxDateTime time)
{
    std::vector<Event> events;
    for(size_t i=0; i<m_events.size(); i++)
    {
        if(!m_events[i].IsReminded()
            && m_events[i].GetReminderDate() <= time)
            events.push_back(m_events[i]);
    }
    return events;
}

void Engine::SetEventReminded(long id, bool reminded)
{
    try
    {
        static const char sql[] =
            "update EVENTS "
            "SET REMINDED = ?"
            "where EVENTID = ?";
        wxSQLite3Statement st = m_db.PrepareStatement(sql);
        st.Bind(1, reminded);
        st.Bind(2, wxLongLong(id));
        st.ExecuteUpdate();
        st.Finalize();
        size_t index = GetEventIndexByID(id);
        if(index != -1)
        {
            m_events[index].SetReminded(reminded);
        }
    }
    catch(const wxSQLite3Exception& e)
    {
        wxLogError(wxString::Format(_("Error in EditEvent: %s"), e.GetMessage().c_str()));
    }
}

bool Engine::HasRecurrence(long id)
{
    for(size_t i=0; i<m_recurrences.size(); i++)
    {
        if(m_recurrences[i].GetEventID() == id)
            return true;
    }
    return false;
}

Recurrence Engine::GetRecurrenceByEventID(long eventid)
{
    for(size_t i=0; i<m_recurrences.size(); i++)
    {
        if(m_recurrences[i].GetEventID() == eventid)
            return m_recurrences[i];
    }
    return Recurrence();
}

Recurrence Engine::AddRecurrence(int type, int days, int months, long eventid)
{
    try
    {
        static const char sql[] =
            "insert into RECURRENCE ("
            "TYPE, DAYS, MONTHS, EVENTID)"
            "values (?, ?, ?, ?)";
        wxSQLite3Statement st = m_db.PrepareStatement(sql);
        st.Bind(1, type);
        st.Bind(2, days);
        st.Bind(3, months);
        st.Bind(4, wxLongLong(eventid));
        st.ExecuteUpdate();
        long id = m_db.GetLastRowId().ToLong();
        st.Finalize();
        Recurrence rec(id, type, days, months, eventid);
        m_recurrences.push_back(rec);
        return rec;
    }
    catch(const wxSQLite3Exception& e)
    {
        wxLogError(wxString::Format(_("Error in AddRecurrence: %s"), e.GetMessage().c_str()));
    }
    return Recurrence();
}

void Engine::RemoveRecurrence(long id)
{
    try
    {
        static const char sql[] = "delete from RECURRENCE where EVENTID = ?";
        wxSQLite3Statement st = m_db.PrepareStatement(sql);
        st.Bind(1, wxLongLong(id));
        st.ExecuteUpdate();
        st.Finalize();
        std::vector<Recurrence>::iterator i;
        for(i = m_recurrences.begin(); i != m_recurrences.end(); ++i)
        {
            if((*i).GetEventID() == id)
            {
                m_recurrences.erase(i);
                break;
            }
        }
    }
    catch(const wxSQLite3Exception& e)
    {
        wxLogError(wxString::Format(_("Error in RemoveRecurrence: %s"), e.GetMessage().c_str()));
    }
}

void Engine::EditRecurrence(long id, int type, int days, int months)
{
        try
    {
        static const char sql[] =
            "update RECURRENCE "
            "set TYPE = ?, DAYS = ?, MONTHS = ?"
            "where EVENTID = ?";
        wxSQLite3Statement st = m_db.PrepareStatement(sql);
        st.Bind(1, type);
        st.Bind(2, days);
        st.Bind(3, months);
        st.Bind(4, wxLongLong(id));
        st.ExecuteUpdate();
        st.Finalize();
        size_t index = GetRecurrenceIndexByEventID(id);
        if(index != -1)
        {
            m_recurrences[index].SetType(type);
            m_recurrences[index].SetDays(days);
            m_recurrences[index].SetMonths(months);
        }
    }
    catch(const wxSQLite3Exception& e)
    {
        wxLogError(wxString::Format(_("Error in EditEvent: %s"), e.GetMessage().c_str()));
    }
}

size_t Engine::GetRecurrenceIndexByEventID(long id)
{
    for(size_t i=0; i<m_recurrences.size(); i++)
    {
        if(m_recurrences[i].GetEventID() == id)
            return i;
    }
    return -1;
}


